(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/WuddySys.sml,v
   Comment: Specialization of Wuddy to support representation of VisualSTATE models
   Release: scope-0-15_final
     
     Copyright (C) 2001-2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTAW.LITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, W.ston, MA 02111-1307, USA.

   THIS MODULE USES IMPERATIVE FEATURES!!!

   WuddySys.sml,v 1.21 2003/07/07 23:53:15 wasowski Exp
*)

structure WuddySys = struct

open Absyn Haskellish Err
infixr 0 $

type 'a env = 'a SymbolEnv.env
type symbol = Symbol.symbol	
type 'a vector = 'a Vector.vector

type posneg = stateid list * stateid list (* pos,neg -> czy na pewno tu? *)

(* STATEFUL *)

val symbol2i_env : int env ref      = ref SymbolEnv.empty 
val i2symbol_env : symbol vector ref = ref (Vector.fromList [])

fun symbol2i (s :symbol) :int = SE.lookup (!symbol2i_env) s  
fun i2symbol (i :int) :symbol = Vector.sub(!i2symbol_env,i)

structure W = Wuddy ( struct type var = symbol 
			     val toInt = symbol2i
			     val fromInt = i2symbol
			     val toString = Symbol.name end )
open W

infix 7 AND
infix 6 OR
infix 4 IMP BIIMP



(* STATEFUL *)

fun init () = bdd.init 10000 10000

fun done () = (

      symbol2i_env := SymbolEnv.empty; 
      i2symbol_env := Vector.fromList [];
      W.done()
)

fun build_var_bindings (es :symbol list) (root :'aor state) :unit = 
				(* events states *)
        let val _ = symbol2i_env := empty
            fun fo (s :'aor orstate,  (i,l)) = (
		symbol2i_env := SE.bind1 (!symbol2i_env) (nameor s,i);
		(i+1,nameor s :: l))
            fun fa (s :'aor andstate, (i,l)) = (
                symbol2i_env := SE.bind1 (!symbol2i_env) (nameand s,i);
                (i+1, nameand s::l))
            val (varnum,l) :int * symbol list = hfold fo fa (0,[]) root
	    fun fe (e,(i,l)) = (
		symbol2i_env := SE.bind1 (!symbol2i_env) (e,i);
		(i+1, e::l)) 
	    val (varnum,l) = foldl fe (varnum,l) es
        in
                i2symbol_env := Vector.fromList (rev l);
                W.realloc varnum
        end 


(* building representation of hierarchy *)
 
fun buildand (s :'aor andstate) =
	 let val e = W.bddVar (nameand s)
	     val kids = map (W.bddVar o nameor) (childrenand s)
	     val kidsand = foldl W.AND W.TRUE kids
	     val kidsor  = foldl W.OR W.FALSE kids
	 in foldl (fn (s,d)=> W.AND(buildor s,d)) 
		(W.AND(W.IMP(e,kidsand),W.IMP(kidsor,e)))
		(childrenand s) 
	 end handle e => intErr e 200212118 
	
and buildor (s :'aor orstate) =
	let val e       = W.bddVar (nameor s)
	    val kids    = map nameand (childrenor s)
	    val kidsxor = generalized_xor kids 
	    val kidsor  = foldl W.OR W.FALSE (map W.bddVar kids)
	in  foldl (fn (s,d) => (W.AND(buildand s,d)))
		(W.AND(W.IMP(e,kidsxor),W.IMP(kidsor,e)))
		(childrenor s)
	end handle e => intErr e 200212117

(* building representation of concurrent state machines -- like
   in flattening *)
(* WARNING! FIXME: currently build_flat assumes that hierarchical system
   had been built prior to call to build_flat. This function does not
   initialize variables in the BDD module (map names to indices) *)

fun build_flat (s :'aor state) =
	let fun fo (s :'aor orstate, phi :bdd) = 
		let val e       = W.bddVar (nameor s)
		    val kids    = map nameand (childrenor s)
		in generalized_xor kids AND phi end
	    fun fa (s :'aor andstate, phi :bdd) = phi
	in hfold fo fa (bddVar (name s)) s end 

(* Used to build parts of the hierarchy                 *)	
(* Also assumes variables to be intialized              *)
(* assumes that variables are already bound to integers *)
fun build_local (s :'aor state) :bdd = 
	let val r = (lift_orand buildor buildand) s
	in (bddVar $ name s) AND r end 
	handle e => intErr e 200212170

(* bind variables and call build_local *)
fun build (s :'aor state) :bdd = 
	let val _ = build_var_bindings [] s 
	in  build_local s end handle e => intErr e 200212113
	handle e => intErr e 200307050

(* this one also adds events to the bdd. Current semantics is:
   exactly one event present at a time. This excludes conjunctions
   of events on transitions and "self-firing" transitions. Both however
   seem to be easily achievable by relaxing the condition *)

fun build_sys (sys :('aor,'atr) system) :bdd =
      let val events = map #1 (SE.toList (#evtypes sys))
          val _      = build_var_bindings events (getroot (#hierarchy sys))  
          val hbdd   = build_local $ InclAND $ gettree $ #hierarchy sys
          val ebdd   = W.generalized_xor events
      in W.AND(hbdd,ebdd) end

fun build_guard ((pos,neg) :posneg) :bdd =
	let val pos1 = map (fn s => (s,true )) pos
	    val neg1 = map (fn s => (s,false)) neg
	    val ass  = assignment (pos1 @ neg1)
			handle e => intErr e 200212116
	in fromAssignment ass end handle e => intErr e 200212115


	




(* FIXME? This function imposes an order on variables in bdd. The order is
   slightly weird as the top variables are leaves of the statecharts. IT
   might be that the tree would become much more regular if the order is
   changed. At the moment this does not influence operation, but it might
   become important for big models. AW 05/07/2002 *)

(* TODO: add events *)

(*  local  fun build_var_bindings (sys :system) : symbol array * int env =
		let val root = getroot (#hierarchy sys)
		    fun fo (s :orstate,  l) = nameor s::l
		    fun fa (s :andstate, l) = nameand s::l
		    val l = rev (hfold fo fa [] root)
		    val n = List.tabulate (length l, fn x=>x)
		    val a = Array.fromList l
		    val e = bindZip SymbolEnv.empty (l,n)
		in (a,e) end 
	
	(*fun build_var_bindings (sys :system) : symbol array * int array =
		let val maxsy = Symbol.toInt (Symbol.fresh())
		    val root  = getroot (#hierarchy sys)
		    fun fo (s :orstate,  l) = nameor s::l
		    fun fa (s :andstate, l) = nameand s::l
		    val l = rev (hfold fo fa [] root)
		    val a = Array.fromList l
		    val n1 = List.tabulate(length l, fn x => x)
		    val e1 = ListPair.foldl 
				(fn (i,j,e) => Intmap.insert (e,i,j))
				(Intmap.empty()) 
				(map Symbol.toInt l,n1)
		    val n = List.tabulate (maxsy,
				fn i => (Intmap.retrieve(e1,i) handle NotFound => ~1))
		    val e = Array.fromList n
		in (a,e) end


	fun test () =
		let val is = List.tabulate ( Array.length (!i2symbol_env), fn x => x )
		    fun f i = if (symbol2i o i2symbol) i <> i then raise Fail "bdvar"
			      else ()
		in app f is end
	*)
  in 
	fun use (sys :system) =  
		let val (i2s,s2i) = build_var_bindings sys
		in 
			i2symbol_env := i2s; 
			symbol2i_env := s2i;
			W.realloc (Array.length i2s);
			(* test() *)
			sysbdd := build_sys sys
		end handle e => intErr e 200212119 
  end *)

(* -- following functions are concerned with representing
      transitions in BDD *)

(* builds representation for transition trigger. Allows both
     simple and disjunction (group) triggers *)

fun build_trigger (sys :('aor,'atr) system) (t :'atr tran) :bdd =
	case #trigger t of
		EventGroup g  => let val es = Set.listItems 
						(SE.lookup (#evgroups sys) g)
				 in foldl W.OR FALSE (map bddVar es) end
	| Event (e,_) => bddVar e

(* builds a representation of transitions condition part.
   STATEFUL Assumes that the transition only refers to symbols 
   with translations defined in current environment. *)

(* TODO: I start to believe that making this stateful was a mistake.
   Perhaps translator should passed to functions as parameter *)

fun build_trancond (sys :('aor,'atr) system) (t :'atr tran) :bdd =
	build_trigger sys t AND build_guard (#pos t,#neg t)


(* get one satisfying assignment and exclude it from bdd. See comment
   by allsat for explanation of vars param. vars is assumed to contain
   no repetitions *)

fun nextsat (r :bdd) (vars :stateid list) 
		:(stateid list * stateid list) list * bdd =
	let val ass = W.satone r
	    val r1  = W.AND(W.NOT(W.fromAssignment ass),r)
	    val (pos,neg) = List.partition #2 (W.getAssignment ass)
	    val pos1 = map #1 pos
	    val neg1 = map #1 neg
	    fun fresh s = List.all (fn s1 => s1 <> s) pos1 andalso
			  List.all (fn s1 => s1 <> s) neg1 
	    val fvars = List.filter fresh vars
	    fun extend s ((p,n),ass) = (p,s::n)::(s::p,n)::ass
	    fun mult []      ass = ass
	      | mult (s::ta) ass = mult ta (foldl (extend s) [] ass)
	in (mult fvars [(pos1,neg1)],r1) end

(* vars is the set of variables which *must* be included in solutions
     returned. All assignements returned refer to vars union support(r). 
     To get result on support only use allsat r []. The overhead of processing 
     empty list is very small. vars may contain repetitions which will be
     filtered out. 
     
     FIXME: this is very inefficient. Joern has better version of Buddy which
     includes AllSat. I am not sure however if he had not implemented this in
     the same ugly way. If not it is worth to be ported. Otherwise Ken
     suggested better ways of searching for Paths to truth.
*)

fun allsat (r :bdd) (vars :stateid list) :(stateid list * stateid list) list =
	let val vars1 = Set.listItems $ Set.fromList vars 
	    fun f (r :bdd) acc =
		if equal r FALSE then acc
		else let val (lst,r1) = nextsat r vars1 in f r1 (lst @ acc) end
	in f r [] end

(* guard reconstruction algorithm EXPERIMENTAL *)
(* h describes valid state configurations, g describes the guard itself  *)

fun get_guard (h :bdd) (g :bdd) :posneg list =
	let val psi = h IMP g
	    val psi1= simplify psi h
	    val dnf = allsat psi1 []
	in dnf end (* FIXME: poorman's solution *)

		

end
