(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/acfg/acfg-builder-sig.sml,v
   Comment: 
   Release: scope-0-15_final

     Copyright (C) 2001-2002 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   acfg-builder-sig.sml,v 1.6 2003/01/28 08:11:38 wasowski Exp 
*)

signature ACFG_BUILDER = sig

        include ACFG

	type ''b rule = bdd.bdd * ''b
	type (''a,'c) splitdata = bdd.bdd * ''a * 'c * 'c
	
	type (''b,'c) partition = bdd.bdd * ''b rule list * 'c -> 
				(''b rule list * 'c) list

	type (''a,''b,'c) pull_up = bdd.bdd * ''b rule list * 'c -> 
				(''a,'c) splitdata

	val build_acfg : (''b,'c) partition -> (''a,''b,'c) pull_up ->
				bdd.bdd -> ''b rule list -> 'c -> (''a,''b) acfg 


end

(*
	The convention for type variables:
	 ''a denotes type of conditions,
	 ''b denotes type of actions,
	 ''c denotes type of state of computation of heuristics 

	[''b rule] is the type of guarded execution rules. 
	Execution rule is a condition and an action. Condition is represented
	by the bdd. Action can be arbitrary equality type.

	[(''a,'c)] is the type of results of pull_up function (see below).
	Roughyl it represents the condition on which a search space is split
	and two states of further computation, one for low branch, one for
	high branch. The condition is represented both logically (in bdd) and
	syntactically (as element of ''a). These two are supposed to be
	equivalent.

	[(''b,'c) partition] type of partitioning heuristics, i.e functions
	which break current sets of rules into several sets, which should be
	processed independently (in sequence). The blocks are represented as
	formulas. A new state of computation is returned for each block. The
	generic builder uses formulas returned by partition to compute actual
	blocks by conjuncting the block formula to all rules and filtering
	out unsatisfied rules.  Finally empty blocks are filtered out.
	Partition should never return an empty block (even on empty list of
	actions -> then it should return no blocks)

	[(''a,''b,'c) pull_up] type of pull_up heuristics, i.e functions
	selecting a condition to test on for current set of rules.

	[build_acfg partition pull_up Fi Lambda e] a function which given a
	partitioning and pull_up heuristics produces an acyclic control flow
	graph for executing given guarded rules Lambda in environment Fi. e
	is the initial state of computation passed to heuristics.
        
*)
