(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/util/Symbol.sml,v
   Comment: Symbol structure for symbol tables
   Release: scope-0-15_final

     Copyright (C) 2001-2002 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   NOTES:

   From A.Appel "Modern Compiler Implementation" Modified by AW. Uses
   Polyhash available in MoscowML (instead of Hashtable as in the book). Also
   restructured in two files: separately for signature and structure in the
   native MoscowML style.

   Symbol.sml,v 1.1 2002/12/11 11:49:17 wasowski Exp
*)

structure Symbol :> Symbol =
struct

  type symbol = string * int

  fun compare (s1 :symbol, s2: symbol) = Int.compare (#2 s1, #2 s2)

  structure H = Polyhash (*HashTable*) 

  exception Symbol


  (* symbol table primitives *)

  val nextsym = ref 0
  val freshid = ref 0
  val sizeHint = 128

  val hashtable : (string,int) H.hash_table ref = 
		ref (H.mkTable(H.hash, op= ) (sizeHint,Symbol))
  
  fun symbol name = ((name, H.find (!hashtable) name)
		    handle Symbol => (let val i = !nextsym in 
					nextsym := i+1;
					H.insert (!hashtable) (name,i);
					(name,i)
				   end))

  fun fresh () = let val i    = !freshid
		     val fsym = "id" ^ Int.toString i 
		 in 
			(H.find (!hashtable) fsym; 
			 freshid := i+1;
			 fresh())
			handle Symbol => symbol fsym
		 end
		 
		  
  
  fun faithful(s,n) = s
  fun obscuring(s,n) = if ( String.isPrefix "_" s ) 
					then s
					else "Symbol" ^ Int.toString n
					

  val name' = ref faithful
  val name  = fn (s,n) => (!name')(s,n)

  fun anonymize anon = if anon	then name' := obscuring
				else name' := faithful
				

  fun toInt (s,n) = n

  val nullsymbol = symbol "";
  
end
